from django.db import models
from django.conf import settings
from django.utils.functional import cached_property
from rest_framework.exceptions import ValidationError
from enumfields import EnumField, Enum
from ponos.models import Artifact
from arkindex_common.enums import DataImportMode
from arkindex_common.ml_tool import MLToolType
from arkindex.project.aws import S3FileMixin, S3FileStatus
from arkindex.project.models import IndexableModel
from arkindex.project.fields import ArrayField
from arkindex.dataimport.providers import git_providers, get_provider
from ponos.models import Workflow, State
from os import path
import shlex
import yaml
import uuid
import urllib.parse


class DataImport(IndexableModel):
    """
    A single import workflow
    """

    creator = models.ForeignKey('users.User', on_delete=models.CASCADE, related_name='imports')
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='imports', null=True)
    mode = EnumField(DataImportMode, max_length=30)
    files = models.ManyToManyField('dataimport.DataFile', related_name='imports')
    revision = models.ForeignKey(
        'dataimport.Revision', related_name='dataimports', on_delete=models.CASCADE, blank=True, null=True)
    workflow = models.OneToOneField('ponos.Workflow', on_delete=models.SET_NULL, null=True, blank=True)
    versions = models.ManyToManyField('dataimport.WorkerVersion', through='dataimport.WorkerRun', related_name='imports')

    # Used to define the root element on which the workflow will run
    element = models.ForeignKey('documents.Element', on_delete=models.SET_NULL, null=True, blank=True)

    # Used to define the element type during folder elements creation for Files/IIIF imports
    folder_type = models.ForeignKey(
        'documents.ElementType',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='folder_imports',
    )

    # Used to define the element type during non-folder elements creation for Files/IIIF imports,
    # or to filter elements by type with Elements imports
    element_type = models.ForeignKey(
        'documents.ElementType',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='element_imports',
    )

    # Used to filter out elements with a name that doesn't contain the substring, only on Elements imports
    name_contains = models.CharField(null=True, blank=True, max_length=150)

    # Used to filter elements by best class, either having or not having a best class (boolean)
    # or having a specific best class (UUID), only on Elements imports
    best_class = models.CharField(null=True, blank=True, max_length=150)

    # Used to save a user's selection for Element workflows
    elements = models.ManyToManyField(
        'documents.Element',
        through='dataimport.DataImportElement',
        related_name='imports',
    )

    # Used to include sub-elements of the actual query
    load_children = models.BooleanField(default=False)

    # Used to define the collection ID and entities import for Transkribus import
    collection_id = models.PositiveIntegerField(null=True, blank=True)
    build_entities = models.BooleanField(default=False)

    class Meta:
        ordering = ['corpus', '-created']

    @property
    def state(self):
        if not self.workflow:
            return State.Unscheduled
        # This allows annotating a DataImport queryset with "last_run" and preventing duplicate SQL queries
        if hasattr(self, 'last_run'):
            # last_run may be None when there is a workflow without any tasks
            if self.last_run is None:
                return State.Unscheduled
            return self.workflow.get_state(self.last_run)
        else:
            return self.workflow.state

    def build_workflow(self, chunks=None, thumbnails=False, corpus_id=None):
        '''
        Create a ponos workflow with a recipe according to configuration
        '''
        ml_workflow_chunks = 1
        import_task_name = 'import'
        if self.mode == DataImportMode.Elements:
            import_task_name = 'initialisation'

        assert self.workflow is None, 'A workflow is already setup'
        if self.mode == DataImportMode.Repository and self.revision is not None and not self.revision.repo.enabled:
            raise ValidationError('Git repository does not have any valid credentials')

        if self.mode == DataImportMode.Repository:
            assert self.revision is not None, \
                'A revision is required to create an import workflow from GitLab repository'
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': (
                        f"python -m arkindex_tasks.import_git {self.revision.id}"
                        + (f' --corpus-id {corpus_id}' if corpus_id else '')
                    )
                },
            }

        elif self.mode in (DataImportMode.Elements, DataImportMode.Workers):
            if chunks is not None:
                assert isinstance(chunks, int) and chunks > 0, 'Chunks count should be an positive integer'
                assert chunks < 11, 'Import distribution is limited to 10 chunks'
                ml_workflow_chunks = chunks

            args = [
                'python', '-m', 'arkindex_tasks.init_elements', str(self.id),
                '--chunks-number', str(ml_workflow_chunks),
            ]

            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': ' '.join(args),
                }
            }

        elif self.mode == DataImportMode.IIIF:
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': 'python -m arkindex_tasks.import_iiif.dataimport {}'.format(self.id),
                },
            }

        elif self.mode == DataImportMode.Transkribus:
            args = [
                'python', '-m', 'arkindex_tasks.import_transkribus',
                str(self.collection_id),
                '--corpus', str(self.corpus.id),
            ]
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': ' '.join(args),
                }
            }

        else:
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': 'python -m arkindex_tasks.import_files {}'.format(self.id),
                },
            }

        # Import entities directy after import step
        if self.mode == DataImportMode.Transkribus and self.build_entities:
            transcriptions_path = shlex.quote(path.join('/data', import_task_name, str(self.collection_id), 'transcriptions.json'))
            tasks['build_entities'] = {
                'image': settings.ARKINDEX_TASKS_IMAGE,
                'command': 'python -m arkindex_tasks.build_entities {}'.format(transcriptions_path),
                'parents': [import_task_name],
            }

        elts_chunk_files = ['elements.json']
        if ml_workflow_chunks > 1:
            elts_chunk_files = ['elements_chunk_{}.json'.format(n) for n in range(1, ml_workflow_chunks + 1)]

        for index, elts_chunk in enumerate(elts_chunk_files, start=1):
            # Add a name suffix if task is distributed
            task_suffix = ''
            if len(elts_chunk_files) > 1:
                task_suffix = '_{}'.format(index)

            elements_path = shlex.quote(path.join('/data', import_task_name, elts_chunk))

            # Generate thumbnails directly after import step
            if thumbnails:
                tasks['thumbnails{}'.format(task_suffix)] = {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': 'python3 -m arkindex_tasks.generate_thumbnails {}'.format(elements_path),
                    'parents': [import_task_name],
                }

            # Generate a task for each WorkerRun on the DataImport
            for worker_run in self.worker_runs.all():
                task_name = f'{worker_run.version.slug}{task_suffix}'
                tasks[task_name] = worker_run.build_task_recipe(import_task_name, elements_path, suffix=task_suffix)

        # Build the workflow in db
        recipe = settings.PONOS_RECIPE.copy()
        recipe['tasks'] = tasks
        if 'env' not in recipe:
            recipe['env'] = {}
        recipe['env'].update({
            'ARKINDEX_PROCESS_ID': str(self.id),
        })
        if self.mode == DataImportMode.Transkribus:
            recipe['env']['TRANSKRIBUS_EMAIL'] = settings.TRANSKRIBUS_EMAIL
            recipe['env']['TRANSKRIBUS_PASSWORD'] = settings.TRANSKRIBUS_PASSWORD
        return Workflow.objects.create(recipe=yaml.dump(recipe))

    def start(self, chunks=None, thumbnails=False, corpus_id=None):
        self.workflow = self.build_workflow(chunks, thumbnails, corpus_id)
        self.save()
        # Start the associated workflow
        self.workflow.start()

    def retry(self):
        if self.mode == DataImportMode.Repository and self.revision is not None and not self.revision.repo.enabled:
            raise ValidationError('Git repository does not have any valid credentials')
        if self.workflow:
            self.workflow.retry()
        else:
            self.start()


class DataImportElement(models.Model):
    """
    Link between DataImports and Elements to allow workflows from selections.
    May be removed later once the ElementPath algorithms are optimized,
    as a selection could then be turned into a set.
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    dataimport = models.ForeignKey(DataImport, on_delete=models.CASCADE, related_name='dataimport_elements')
    element = models.ForeignKey('documents.Element', on_delete=models.CASCADE, related_name='dataimport_elements')

    class Meta:
        unique_together = (
            ('dataimport', 'element'),
        )


class DataFile(S3FileMixin, models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255)
    size = models.PositiveIntegerField(help_text='file size in bytes')
    content_type = models.CharField(max_length=50)
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='files')
    status = EnumField(S3FileStatus, default=S3FileStatus.Unchecked, max_length=50)

    s3_bucket = settings.AWS_STAGING_BUCKET

    @cached_property
    def s3_key(self):
        return str(self.id)

    class Meta:
        ordering = ['corpus', 'name']

    def perform_check(self, save=True, raise_exc=False):
        """
        Check Datafile's existence, content type and size from S3
        """
        try:
            assert self.exists(), 'File has not been uploaded'
            assert self.s3_object.content_length == self.size, \
                f'Uploaded file size is {self.s3_object.content_length} bytes, expected {self.size} bytes'
            assert self.s3_object.content_type == self.content_type, \
                f'Uploaded file content type is {self.s3_object.content_type}, expected {self.content_type}'
            self.status = S3FileStatus.Checked
        except AssertionError:
            self.status = S3FileStatus.Error
            if raise_exc:
                raise
        finally:
            if save:
                self.save()


class RepositoryType(Enum):
    IIIF = 'iiif'
    Worker = 'worker'


class Repository(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    url = models.URLField(unique=True)
    type = EnumField(RepositoryType)
    hook_token = models.CharField(max_length=250, unique=True)
    credentials = models.ForeignKey(
        'users.OAuthCredentials', on_delete=models.SET_NULL, related_name='repos', blank=True, null=True)
    provider_name = models.CharField(
        max_length=50,
        choices=[(p.__name__, p.display_name) for p in git_providers],
    )
    git_ref_revisions = models.ManyToManyField('dataimport.Revision', through='dataimport.GitRef')

    class Meta:
        verbose_name_plural = 'repositories'

    @property
    def provider_class(self):
        return get_provider(self.provider_name)

    @property
    def provider(self):
        return self.provider_class(credentials=self.credentials)

    @property
    def enabled(self):
        return self.credentials is not None and self.provider_class is not None

    def __str__(self):
        return f'{self.url}'


class GitRefType(Enum):
    Branch = 'branch'
    Tag = 'tag'


class GitRef(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    type = EnumField(GitRefType)
    name = models.CharField(max_length=250)
    repository = models.ForeignKey('dataimport.Repository', on_delete=models.CASCADE, related_name='refs')
    revision = models.ForeignKey('dataimport.Revision', on_delete=models.CASCADE, related_name='refs')

    class Meta:
        unique_together = (('name', 'repository'), )


class Revision(IndexableModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    repo = models.ForeignKey('dataimport.Repository', on_delete=models.CASCADE, related_name='revisions')
    hash = models.CharField(max_length=50)
    message = models.TextField()
    author = models.CharField(max_length=50)

    class Meta:
        unique_together = (('repo', 'hash'), )

    @property
    def commit_url(self):
        return '{}/commit/{}'.format(self.repo.url.rstrip('/'), self.hash)

    def __str__(self):
        return '{} "{}" by {}'.format(self.hash[:8], self.message.splitlines()[0], self.author)

    @property
    def state(self):
        """
        Computes revision state according to its versions'

        If there is one version in error, revision state is too
        Else if there is one version processing, revision state is too
        Else if all versions are available, then the revision is too
        Else, the revision is created since it has either no version or versions mixing processing/created states
        """
        # This prevents performing another SQL request when versions have already been prefetched.
        # See https://stackoverflow.com/a/19651840/5990435
        if (
            hasattr(self, "_prefetched_objects_cache")
            and self.versions.field.remote_field.get_cache_name()
            in self._prefetched_objects_cache
        ):
            states = set(version.state for version in self.versions.all())
        else:
            states = set(self.versions.values_list('state', flat=True))

        if WorkerVersionState.Error in states:
            return WorkerVersionState.Error
        elif WorkerVersionState.Processing in states:
            return WorkerVersionState.Processing
        elif states == {WorkerVersionState.Available}:
            return WorkerVersionState.Available
        else:
            return WorkerVersionState.Created


class Worker(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=100)
    slug = models.CharField(max_length=100)
    type = EnumField(MLToolType, max_length=50)
    repository = models.ForeignKey('dataimport.Repository', on_delete=models.CASCADE, related_name='workers')

    class Meta:
        unique_together = (('slug', 'repository'),)

    def __str__(self):
        return f'{self.name}'


class WorkerVersionState(Enum):
    Created = 'created'
    Processing = 'processing'
    Available = 'available'
    Error = 'error'


class WorkerVersion(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    worker = models.ForeignKey('dataimport.Worker', on_delete=models.CASCADE)
    revision = models.ForeignKey('dataimport.Revision', on_delete=models.CASCADE, related_name='versions')
    configuration = models.JSONField()
    state = EnumField(WorkerVersionState, default=WorkerVersionState.Created)

    # Foreign key towards the ponos Artifact holding the image content
    docker_image = models.ForeignKey(Artifact, on_delete=models.SET_NULL, null=True, blank=True)

    # The Docker internal image id (sha256:xxx) that can be shared across multiple images
    docker_image_iid = models.CharField(null=True, blank=True, max_length=80)

    class Meta:
        unique_together = (('worker', 'revision'),)
        constraints = [
            models.CheckConstraint(
                check=~models.Q(state=WorkerVersionState.Available, docker_image_id=None),
                name='workerversion_available_requires_docker_image',
            )
        ]

    def __str__(self):
        return f'{self.worker} for revision {self.revision}'

    @property
    def docker_image_name(self):
        parsed_url = urllib.parse.urlparse(self.worker.repository.url)
        return f'{parsed_url.netloc}{parsed_url.path}/{self.worker.slug}:{self.id}'.lower()

    @property
    def slug(self):
        truncated_hash = str(self.id)[0:6]
        return f'{self.worker.slug}_{truncated_hash}'

    @property
    def docker_command(self):
        if 'docker' not in self.configuration:
            return
        return self.configuration['docker'].get('command')


class WorkerRun(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    dataimport = models.ForeignKey('dataimport.DataImport', on_delete=models.CASCADE, related_name='worker_runs')
    version = models.ForeignKey('dataimport.WorkerVersion', on_delete=models.CASCADE, related_name='worker_runs')
    parents = ArrayField(models.UUIDField())

    class Meta:
        unique_together = (('version', 'dataimport'),)

    def build_task_recipe(self, import_task_name, elements_path, suffix=''):
        '''
        Build recipe dict for the Task that will represent this WorkerRun in ponos using :
        - the docker image name given by the WorkerVersion
        - the command given by the WorkerVersion
        - the artifact given by the WorkerVersion that will help us download the distant Docker image (S3)
        - a list of parent tasks slugs, using the base import task as a default
        '''
        if self.parents:
            parents = [
                f'{worker_run.version.slug}{suffix}'
                for worker_run in WorkerRun.objects.filter(id__in=self.parents)
            ]
        else:
            parents = [import_task_name]

        return {
            'image': self.version.docker_image_iid or self.version.docker_image_name,
            'command': self.version.docker_command,
            'artifact': str(self.version.docker_image.id),
            'parents': parents,
            'env': {"TASK_ELEMENTS": elements_path, "WORKER_VERSION_ID": str(self.version.id)}
        }
