from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from arkindex.dataimport.models import Worker, WorkerVersion, WorkerVersionState, Repository, RepositoryType
from arkindex.project.serializer_fields import EnumField
from arkindex_common.ml_tool import MLToolType
from arkindex.dataimport.serializers.git import RevisionWithRefsSerializer
import urllib


class WorkerSerializer(serializers.ModelSerializer):
    """
    Serialize a repository worker
    """
    type = EnumField(MLToolType)

    class Meta:
        model = Worker
        fields = (
            'id',
            'name',
            'slug',
            'type',
        )


class WorkerVersionSerializer(serializers.ModelSerializer):
    """
    Serialize a worker version
    """
    state = EnumField(WorkerVersionState, required=False)
    worker = WorkerSerializer(read_only=True)
    # ModelSerializer does not yet support Django 3.1's JSONField
    # https://github.com/encode/django-rest-framework/pull/7467
    configuration = serializers.JSONField()

    class Meta:
        model = WorkerVersion
        fields = (
            'id',
            'configuration',
            'revision',
            'docker_image',
            'docker_image_iid',
            'docker_image_name',
            'state',
            'worker',
        )
        read_only_fields = ('docker_image_name',)
        # Avoid loading all revisions and all Ponos artifacts when opening this endpoint in a browser
        extra_kwargs = {
            'revision': {'style': {'base_template': 'input.html'}},
            'docker_image': {'style': {'base_template': 'input.html'}},
        }

    def to_representation(self, instance):
        self.fields['revision'] = RevisionWithRefsSerializer(read_only=True)
        return super(WorkerVersionSerializer, self).to_representation(instance)

    def validate(self, data):
        if self.instance:
            # When doing a partial update (PATCH), undefined fields should fall back to the instance's fields
            state = data['state'] if 'state' in data else self.instance.state
            docker_image = data['docker_image'] if 'docker_image' in data else self.instance.docker_image_id
            docker_image_iid = data['docker_image_iid'] if 'docker_image_iid' in data else self.instance.docker_image_iid
        else:
            state = data.get('state')
            docker_image = data.get('docker_image')
            docker_image_iid = data.get('docker_image_iid')

        if state == WorkerVersionState.Available and (not docker_image or not docker_image_iid):
            raise ValidationError([
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ])

        return data


class RepositorySerializer(serializers.ModelSerializer):
    """
    Serialize a repository
    """
    enabled = serializers.BooleanField(read_only=True)
    git_clone_url = serializers.SerializerMethodField()
    workers = WorkerSerializer(many=True, read_only=True)
    type = EnumField(RepositoryType)

    def get_git_clone_url(self, repository):
        # This check avoid to set git_clone_url when this serializer is used to list multiple
        # repositories because self.instance value would be a list, even if the request user is internal
        if self.context['request'].user.is_internal and isinstance(self.instance, Repository) and self.instance.enabled:
            url = urllib.parse.urlparse(self.instance.url)
            return f"https://oauth2:{repository.credentials.token}@{url.netloc}{url.path}"
        return None

    class Meta:
        model = Repository
        fields = (
            'id',
            'url',
            'type',
            'enabled',
            'corpora',
            'git_clone_url',
            'workers',
        )
        extra_kwargs = {
            'id': {'read_only': True},
            'url': {'read_only': True},
            'type': {'read_only': True},
        }
