from django.urls import reverse
from rest_framework import status
from arkindex_common.enums import DataImportMode
from arkindex.dataimport.models import DataImport, DataFile, RepositoryType
from arkindex.documents.models import Corpus, ElementType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import User
from ponos.models import State, Task, Workflow
from unittest.mock import patch, MagicMock
import yaml
import uuid

RECIPE = '''
tasks:
  first:
    image: alpine
  second:
    parents:
      - first
    image: hello-world
'''


class TestImports(FixtureAPITestCase):
    """
    Test data imports management
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(type=RepositoryType.IIIF)
        cls.rev = cls.repo.revisions.get()
        cls.img_df = cls.corpus.files.create(
            name='test.jpg',
            size=42,
            content_type='image/jpg',
        )
        cls.pdf_df = cls.corpus.files.create(
            name='test.pdf',
            size=42,
            content_type='application/pdf',
        )
        cls.iiif_df = cls.corpus.files.create(
            name='test.json',
            size=42,
            content_type='application/json',
        )
        cls.page_type = ElementType.objects.get(corpus=cls.corpus, slug='page')

    def setUp(self):
        super().setUp()
        self.dataimport = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Images)

    def build_task(self, workflow_id, run, state, depth=1):
        """
        Create a simple task with its workflow and process if required
        """
        workflow, created = Workflow.objects.get_or_create(id=workflow_id, defaults={'recipe': RECIPE})
        if created:
            # Creates the associated dataimport
            DataImport.objects.create(workflow=workflow, mode=DataImportMode.Workers, creator=self.user)
        return Task.objects.create(run=run, depth=depth, workflow=workflow, state=state, slug=str(uuid.uuid4()))

    def test_list_requires_login(self):
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        result = data['results'][0]
        self.assertDictEqual(result, {
            'id': str(self.dataimport.id),
            'state': State.Unscheduled.value,
            'mode': DataImportMode.Images.value,
            'corpus': str(self.corpus.id),
            'workflow': None,
        })

    def test_list_workflow_link(self):
        dataimport2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        self.assertIsNone(dataimport2.workflow)

        self.client.force_login(self.user)
        self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport2.id)})
        )
        dataimport2.refresh_from_db()

        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        results = data['results']
        self.assertListEqual(results, [{
            'id': str(dataimport2.id),
            'state': State.Unscheduled.value,
            'mode': DataImportMode.Workers.value,
            'corpus': str(self.corpus.id),
            'workflow': f"http://testserver/ponos/v1/workflow/{dataimport2.workflow.id}/",
        }])

        response = self.client.get(reverse('api:import-list'), {'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        results = data['results']
        self.assertListEqual(results, [{
            'id': str(self.dataimport.id),
            'state': State.Unscheduled.value,
            'mode': DataImportMode.Images.value,
            'corpus': str(self.corpus.id),
            'workflow': None,
        }])

    def test_list_filter_corpus(self):
        self.client.force_login(self.superuser)
        corpus2 = Corpus.objects.create(name='Another corpus', description='something')
        dataimport2 = corpus2.imports.create(creator=self.user, mode=DataImportMode.Images)

        response = self.client.get(reverse('api:import-list'), {'corpus': str(corpus2.id), 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_filter_corpus_invalid(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'corpus': 'oh-no'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'corpus': "'oh-no' is not a valid UUID"})

    def test_list_filter_mode(self):
        self.client.force_login(self.user)
        dataimport2 = self.corpus.imports.create(
            creator=self.user,
            mode=DataImportMode.PDF,
        )
        response = self.client.get(reverse('api:import-list'), {'mode': DataImportMode.PDF.value, 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_filter_unexisting_mode(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'mode': 'unexisting_mode'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'mode': "Mode 'unexisting_mode' does not exist"})

    def test_list_filter_creator_by_email(self):
        self.client.force_login(self.user)
        self.corpus.imports.create(
            creator=User.objects.create_user(email='another_user@user.fr'),
            mode=DataImportMode.PDF,
        )
        response = self.client.get(reverse('api:import-list'), {'creator': 'user@user.fr', 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(self.dataimport.id))

    def test_list_filter_creator_by_id(self):
        self.client.force_login(self.user)
        self.corpus.imports.create(
            creator=User.objects.create_user(email='another_user@user.fr'),
            mode=DataImportMode.PDF,
        )
        response = self.client.get(reverse('api:import-list'), {'creator': self.user.id, 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(self.dataimport.id))

    def test_list_filter_unexisting_creator(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'creator': 'blabla@blabla.fr'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'creator': "User with email or id 'blabla@blabla.fr' does not exist"})

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'creator': '5'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'creator': "User with email or id '5' does not exist"})

    def test_list_filter_id(self):
        self.client.force_login(self.user)

        # Generate a UUID whose first characters are not the same as the first DataImport
        dataimport_id = str(uuid.uuid4())
        while dataimport_id[:10] == str(self.dataimport.id)[:10]:
            dataimport_id = str(uuid.uuid4())

        dataimport2 = self.corpus.imports.create(
            id=dataimport_id,
            creator=self.user,
            mode=DataImportMode.Images,
        )
        response = self.client.get(reverse('api:import-list'), {'id': dataimport_id[:10], 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_no_tasks(self):
        """
        Ensure the DataImport reports an Unscheduled state when there are no tasks in its workflow
        """
        self.assertIsNone(self.dataimport.workflow)
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().delete()

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['state'], State.Unscheduled.value)

    def test_list_process_filter_wrong_state_value(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'state': 'spain'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'state': "State 'spain' does not exist"})

    def test_list_process_state_filter(self):
        """
        Only last run should be used to filter processes by state
        """
        workflows = [uuid.uuid4() for i in range(3)]
        for task in (
            (workflows[0], 0, State.Failed),
            (workflows[0], 1, State.Completed),

            (workflows[1], 0, State.Error),
            (workflows[1], 0, State.Completed),

            (workflows[2], 0, State.Completed),
            (workflows[2], 1, State.Completed),
            (workflows[2], 2, State.Stopped),
        ):
            self.build_task(*task)
        processes = [DataImport.objects.get(workflow_id=w) for w in workflows]
        # Only first dataimport is completed
        self.assertListEqual([p for p in processes if p.state.value == 'completed'], [processes[0]])

        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:import-list'), {'state': 'completed'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual([r['id'] for r in response.json()['results']], [str(processes[0].id)])

    def test_list_process_state_filter_stopped(self):
        workflows = [uuid.uuid4() for i in range(4)]
        for task in (
            # Should be stopping
            (workflows[0], 0, State.Error),
            (workflows[0], 1, State.Stopping),
            # Should be in error state
            (workflows[1], 0, State.Error),
            (workflows[1], 0, State.Stopping),
            # Should be stopped
            (workflows[2], 0, State.Unscheduled),
            (workflows[2], 0, State.Stopped),
            # Should be stopped
            (workflows[3], 0, State.Failed),
            (workflows[3], 1, State.Pending),
            (workflows[3], 1, State.Stopped),
        ):
            self.build_task(*task)
        processes = [DataImport.objects.get(workflow_id=w) for w in workflows]
        self.assertListEqual([p for p in processes if p.state.value == 'stopped'], [processes[2], processes[3]])

        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:import-list'), {'state': 'stopped'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            [r['id'] for r in response.json()['results']],
            [str(processes[2].id), str(processes[3].id)]
        )

    def test_list_process_state_filter_unscheduled(self):
        """
        Processes with no task should be interpreted as unscheduled
        """
        workflow_id = uuid.uuid4()
        for task in (
            (workflow_id, 0, State.Completed),
            (workflow_id, 1, State.Unscheduled)
        ):
            self.build_task(*task)
        processes = [
            DataImport.objects.get(workflow_id=workflow_id),
            DataImport.objects.get(workflow=None)
        ]
        # Wirkflows with no tasks are considered unscheduled
        self.assertListEqual([p for p in processes if p.state.value == 'unscheduled'], processes)

        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:import-list'), {'state': 'unscheduled', 'with_workflow': True})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual([r['id'] for r in response.json()['results']], [str(processes[0].id)])

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:import-list'), {'state': 'unscheduled', 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual([r['id'] for r in response.json()['results']], [str(processes[1].id)])

    def test_details_requires_login(self):
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_details_no_tasks(self):
        """
        Ensure the DataImport reports an Unscheduled state when there are no tasks in its workflow
        """
        self.assertIsNone(self.dataimport.workflow)
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().delete()
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['state'], State.Unscheduled.value)

    def test_details(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.dataimport.id))

    def test_details_destroy_verified(self):
        user = User.objects.create(email='AAAAAAA')
        self.client.force_login(user)
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_details_destroy_running(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        self.dataimport.workflow.tasks.update(state=State.Running)
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ['Cannot edit or delete a workflow while it is running'])

    def test_details_destroy_no_permission(self):
        """
        A user cannot delete a dataimport linked to a corpus he has no write access to
        """
        self.client.force_login(self.user)
        acl_rights = self.corpus.corpus_right.get(user=self.user)
        acl_rights.can_write = False
        acl_rights.can_admin = False
        acl_rights.save()
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': f'You have no write access to corpus "{self.corpus.name}".'})

    def test_details_destroy_no_corpus_forbidden(self):
        """
        A user cannot delete a dataimport with a null corpus if he is not its creator nor an admin
        """
        self.client.force_login(self.user)
        self.assertFalse(self.user.is_admin)
        dataimport = DataImport.objects.create(mode=DataImportMode.Repository, creator=User.objects.create())
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You are not the creator of this process nor an admin.'})

    def test_details_destroy_no_corpus_is_creator(self):
        """
        A user is allowed to delete a dataimport if he is the creator
        """
        self.client.force_login(self.user)
        self.assertFalse(self.user.is_admin)
        dataimport = DataImport.objects.create(mode=DataImportMode.Repository, creator=self.user)
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_details_destroy_no_corpus_is_admin(self):
        """
        A user is allowed to delete a dataimport he has not created if he is an admin
        """
        user = User.objects.create(email='admin', is_admin=True)
        self.client.force_login(user)
        dataimport = DataImport.objects.create(mode=DataImportMode.Repository, creator=User.objects.create())
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_update_process_requires_login(self):
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_update_process_requires_verified(self):
        user = User.objects.create(email="unchecked")
        self.client.force_login(user)
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_update_process_running(self):
        """
        A running process cannot be updated
        """
        self.client.force_login(self.user)
        self.dataimport.start()
        self.dataimport.workflow.tasks.update(state=State.Running)
        response = self.client.patch(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ['Cannot edit or delete a workflow while it is running'])

    def test_update_process_no_permission(self):
        """
        A user cannot update a dataimport linked to a corpus he has no write access to
        """
        self.client.force_login(self.user)
        acl_rights = self.corpus.corpus_right.get(user=self.user)
        acl_rights.can_write = False
        acl_rights.can_admin = False
        acl_rights.save()
        response = self.client.patch(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': f'You have no write access to corpus "{self.corpus.name}".'})

    def test_update_process_no_corpus_forbidden(self):
        self.client.force_login(self.user)
        self.assertFalse(self.user.is_admin)
        dataimport = DataImport.objects.create(mode=DataImportMode.Repository, creator=User.objects.create())
        response = self.client.patch(reverse('api:import-details', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You are not the creator of this process nor an admin.'})

    def test_update_process_wrong_mode(self):
        """
        Other process modes than Workers cannot be patched
        """
        self.client.force_login(self.user)
        forbidden_modes = list(DataImportMode)
        forbidden_modes.remove(DataImportMode.Workers)
        for mode in forbidden_modes:
            dataimport = DataImport.objects.create(mode=mode, creator=self.user)
            response = self.client.patch(reverse('api:import-details', kwargs={'pk': dataimport.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
            self.assertDictEqual(
                response.json(),
                {'detail': 'Only processes of mode DataImportMode.Workers can be updated'}
            )

    def test_update_process_corpus_no_write_right(self):
        self.client.force_login(self.user)
        acl_rights = self.corpus.corpus_right.get(user=self.user)
        acl_rights.can_write = False
        acl_rights.can_admin = False
        acl_rights.save()
        dataimport = DataImport.objects.create(mode=DataImportMode.Workers, corpus=self.corpus, creator=self.user)
        response = self.client.patch(reverse('api:import-details', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': f'You have no write access to corpus "{self.corpus.name}".'}
        )

    def test_update_process(self):
        self.client.force_login(self.user)
        dataimport = DataImport.objects.create(mode=DataImportMode.Workers, corpus=self.corpus, creator=self.user)
        self.assertTupleEqual(
            (dataimport.name_contains, dataimport.element_type, dataimport.load_children),
            (None, None, False)
        )
        response = self.client.patch(
            reverse('api:import-details', kwargs={'pk': dataimport.id}),
            {
                'name_contains': 'AAA',
                'element_type': 'page',
                'load_children': True
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        dataimport.refresh_from_db()
        self.assertTupleEqual(
            (dataimport.name_contains, dataimport.element_type, dataimport.load_children),
            ('AAA', self.page_type, True)
        )

    def test_update_process_reset_fields(self):
        self.client.force_login(self.user)
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Workers,
            corpus=self.corpus,
            creator=self.user,
            name_contains='AAA',
            element_type=self.page_type,
            load_children=True
        )
        response = self.client.patch(
            reverse('api:import-details', kwargs={'pk': dataimport.id}),
            {
                'name_contains': None,
                'element_type': None,
                'load_children': False
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        dataimport.refresh_from_db()
        self.assertTupleEqual(
            (dataimport.name_contains, dataimport.element_type, dataimport.load_children),
            (None, None, False)
        )

    def test_update_process_wrong_type(self):
        self.client.force_login(self.user)
        dataimport = DataImport.objects.create(mode=DataImportMode.Workers, corpus=self.corpus, creator=self.user)
        response = self.client.patch(
            reverse('api:import-details', kwargs={'pk': dataimport.id}),
            {'element_type': 'unexisting_type'},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'element_type': ['Object with slug=unexisting_type does not exist.']}
        )

    def test_update_process_fields(self):
        """
        Only name_contains, element_type and load_children values can be updated
        """
        self.client.force_login(self.user)
        dataimport = DataImport.objects.create(mode=DataImportMode.Workers, corpus=self.corpus, creator=self.user)
        response = self.client.patch(
            reverse('api:import-details', kwargs={'pk': dataimport.id}),
            {
                'id': str(uuid.uuid4()),
                'revision': str(uuid.uuid4()),
                'element': str(uuid.uuid4()),
                'workflow': str(uuid.uuid4()),
                'creator': str(uuid.uuid4()),
                'files': [1, 2, 3],
                'mode': 'iiif',
                'folder_type': 'register',
                'state': 'success',
                'name_contains': 'AAA',
                'element_type': 'page',
                'load_children': True
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            'name_contains': 'AAA',
            'element_type': 'page',
            'load_children': True,
            'id': str(dataimport.id),
            'corpus': str(self.corpus.id),
            'element': None,
            'files': [],
            'mode': 'workers',
            'revision': None,
            'state': 'unscheduled',
            'workflow': None
        })

    def test_retry_requires_login(self):
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retry_only_final(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        self.assertEqual(self.dataimport.state, State.Unscheduled)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retry_running(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        task = self.dataimport.workflow.tasks.all().last()
        task.state = State.Running
        task.save()
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retry(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.dataimport.state, State.Error)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataimport.refresh_from_db()
        self.assertEqual(self.dataimport.state, State.Unscheduled)

    def test_retry_no_workflow(self):
        self.client.force_login(self.user)
        self.assertIsNone(self.dataimport.workflow)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataimport.refresh_from_db()
        self.assertEqual(self.dataimport.state, State.Unscheduled)
        self.assertIsNotNone(self.dataimport.workflow)

    def test_retry_repo_disabled(self):
        self.client.force_login(self.user)
        self.dataimport.mode = DataImportMode.Repository
        self.dataimport.revision = self.rev
        self.dataimport.save()
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.dataimport.state, State.Error)
        self.creds.delete()
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.creds.save()

    def test_retry_no_corpus_not_creator_nor_admin(self):
        """
        A user that is not the creator nor admin cannot restart a dataimport that is not linked to any corpus
        """
        self.client.force_login(self.user)
        self.assertFalse(self.user.is_admin)
        creator = User.objects.create()
        dataimport = DataImport.objects.create(mode=DataImportMode.Repository, revision=self.rev, creator=creator)
        dataimport.start()
        dataimport.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(dataimport.state, State.Error)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_from_files_requires_login(self):
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_from_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.Images)
        self.assertListEqual(list(dataimport.files.all()), [self.img_df])
        self.assertEqual(dataimport.folder_type.slug, 'volume')
        self.assertEqual(dataimport.element_type.slug, 'page')
        self.assertIsNone(dataimport.element)

    def test_from_files_pdf(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'mode': 'pdf',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.PDF)
        self.assertListEqual(list(dataimport.files.all()), [self.pdf_df])
        self.assertIsNone(dataimport.element)

    def test_from_files_iiif(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.IIIF)
        self.assertListEqual(list(dataimport.files.all()), [self.iiif_df])
        self.assertEqual(dataimport.folder_type.slug, 'volume')
        self.assertEqual(dataimport.element_type.slug, 'page')
        self.assertIsNone(dataimport.element)

    def test_from_files_iiif_requires_folder_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['IIIF imports require both folder_type and element_type']
        })

    def test_from_files_invalid_mode(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'repository',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_pdf_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'mode': 'pdf',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_pdf_single(self):
        self.client.force_login(self.user)
        pdf2 = DataFile.objects.create(
            name='test2.pdf', size=1337, content_type='application/pdf', corpus=self.corpus)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id), str(pdf2.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'pdf',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_images_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_iiif_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'iiif',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_folder_id(self):
        self.client.force_login(self.user)
        volume = self.corpus.elements.get(name='Volume 1')
        response = self.client.post(
            reverse('api:import-from-files'),
            {
                'files': [str(self.pdf_df.id)],
                'mode': 'pdf',
                'folder_id': str(volume.id),
                'element_type': 'page',
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.element, volume)

    def test_from_files_no_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_start_process_requires_login(self):
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(self.dataimport.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_start_process_invalid_dataimport_id(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_start_process_wrong_dataimport_mode(self):
        self.assertNotEqual(self.dataimport.mode, DataImportMode.Workers)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(self.dataimport.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ['Only a DataImport with Workers mode and not already launched can be started later on'])

    def test_start_process_dataimport_already_started(self):
        dataimport2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        dataimport2.workflow = Workflow.objects.create(recipe=RECIPE)
        dataimport2.save()
        self.assertIsNotNone(dataimport2.workflow)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ['Only a DataImport with Workers mode and not already launched can be started later on'])

    def test_start_process(self):
        dataimport2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        self.assertIsNone(dataimport2.workflow)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['id'], str(dataimport2.id))
        dataimport2.refresh_from_db()
        self.assertEqual(dataimport2.state, State.Unscheduled)
        self.assertIsNotNone(dataimport2.workflow)

    def test_start_process_params_validation(self):
        dataimport = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        self.client.force_login(self.user)
        wrong_params_checks = [
            ({'chunks': 0}, {'chunks': ['Ensure this value is greater than or equal to 1.']}),
            ({'chunks': 20}, {'chunks': ['Ensure this value is less than or equal to 10.']}),
            ({'chunks': 'max'}, {'chunks': ['A valid integer is required.']}),
            ({'thumbnails': 'gloubiboulga'}, {'thumbnails': ['Must be a valid boolean.']})
        ]
        for (params, check) in wrong_params_checks:
            response = self.client.post(reverse('api:process-start', kwargs={'pk': str(dataimport.id)}), params)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(response.json(), check)

    @patch('arkindex.dataimport.models.DataImport.worker_runs')
    def test_start_process_parameters(self, worker_runs_mock):
        """
        It should be possible to pass chunks and thumbnails parameters when starting a workflow
        """
        dataimport = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        # Add a worker run to this process
        run_mock = MagicMock()
        run_mock.version.slug = 'my_worker'
        run_mock.build_task_recipe.return_value = {'image': ''}
        worker_runs_mock.all.return_value = [run_mock]

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport.id)}),
            {'chunks': 3, 'thumbnails': 'true'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        dataimport.refresh_from_db()
        self.maxDiff = None
        recipe_dump = yaml.safe_load(dataimport.workflow.recipe)
        self.assertCountEqual(
            recipe_dump['tasks'].keys(),
            ['import', 'my_worker_1', 'my_worker_2', 'my_worker_3', 'thumbnails_1', 'thumbnails_2', 'thumbnails_3']
        )
