from django.urls import reverse
from rest_framework import status
from arkindex_common.ml_tool import MLToolType
from arkindex.dataimport.models import Repository, Revision, Worker, WorkerVersion, WorkerVersionState, RepositoryType
from arkindex.project.tests import FixtureAPITestCase
from ponos.models import Workflow

RECIPE = '''
tasks:
  first:
    image: alpine
  second:
    parents:
      - first
    image: hello-world
'''


class TestWorkersWorkerVersions(FixtureAPITestCase):
    """
    Test workers and worker versions endpoints
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(type=RepositoryType.Worker)
        cls.rev = cls.repo.revisions.get()

        cls.worker_1 = Worker.objects.get(slug='reco')
        cls.worker_2 = Worker.objects.get(slug='dla')
        cls.rev2 = Revision.objects.create(
            hash='1234',
            message='commit message',
            author='Bob',
            repo=cls.repo,
        )

    def setUp(self):
        self.version_1 = WorkerVersion.objects.get(worker=self.worker_1)

    # Tests on get_query_set for WorkerList
    def test_workers_list_requires_login(self):
        response = self.client.get(reverse('api:repository-workers', kwargs={'pk': str(self.repo.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_workers_list(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:repository-workers', kwargs={'pk': str(self.repo.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data), 2)
        self.assertCountEqual(data, [
            {
                'id': str(self.worker_1.id),
                'name': 'Recognizer',
                'slug': 'reco',
                'type': 'recognizer'
            }, {
                'id': str(self.worker_2.id),
                'name': 'Document layout analyser',
                'slug': 'dla',
                'type': 'dla'
            }
        ])

    def test_workers_list_filter_repository(self):
        self.client.force_login(self.user)
        repo2 = Repository.objects.create(
            url='http://gitlab/repo2',
            type=RepositoryType.Worker,
            hook_token='hook-token2',
            credentials=self.creds,
            provider_name='GitLabProvider'
        )
        worker_2 = Worker.objects.create(
            repository=repo2,
            name='Worker 2',
            slug='worker_2',
            type=MLToolType.Classifier
        )

        response = self.client.get(reverse('api:repository-workers', kwargs={'pk': str(repo2.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data), 1)
        worker = data[0]
        self.assertEqual(worker['id'], str(worker_2.id))

    # Tests on get_query_set for WorkerRetrieve
    def test_workers_retrieve(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:worker-retrieve', kwargs={'pk': str(self.worker_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.worker_1.id),
            'name': self.worker_1.name,
            'slug': self.worker_1.slug,
            'type': self.worker_1.type.value
        })

    def test_workers_retrieve_required_login(self):
        response = self.client.get(reverse('api:worker-retrieve', kwargs={'pk': str(self.worker_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    # Test on post for WorkerList
    def test_workers_post_requires_login(self):
        response = self.client.post(
            reverse('api:repository-workers', kwargs={'pk': str(self.repo.id)}),
            data={'name': 'Worker post', 'slug': 'worker_post', 'type': 'classifier'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_workers_post_invalid_repository(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:repository-workers', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}),
            data={'name': 'Worker post', 'slug': 'worker_post', 'type': 'classifier'}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_workers_post_create_worker(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:repository-workers', kwargs={'pk': str(self.repo.id)}),
            data={'name': 'Worker post', 'slug': 'worker_post', 'type': 'classifier'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertNotEqual(data['id'], str(self.worker_1.id))
        self.assertEqual(data['name'], 'Worker post')
        self.assertEqual(data['slug'], 'worker_post')
        self.assertEqual(data['type'], MLToolType.Classifier.value)

    def test_workers_post_return_existing_worker(self):
        self.client.force_login(self.user)
        # A worker named Worker 1 already exists with the same slug on this repo
        response = self.client.post(
            reverse('api:repository-workers', kwargs={'pk': str(self.repo.id)}),
            data={'name': 'Worker Test', 'slug': 'reco', 'type': 'classifier'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.worker_1.id))
        self.assertEqual(data['name'], 'Recognizer')
        self.assertEqual(data['slug'], 'reco')
        self.assertEqual(data['type'], MLToolType.Recognizer.value)

    def test_workers_post_empty(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:repository-workers', kwargs={'pk': str(self.repo.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'name': ['This field is required.'],
            'slug': ['This field is required.'],
            'type': ['This field is required.']
        })

    # Tests on get_query_set for WorkerVersionList
    def test_versions_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_versions_list(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data), 1)
        version = data[0]
        self.assertEqual(version['id'], str(self.version_1.id))
        self.assertEqual(version['configuration'], {"test": 42})
        self.assertEqual(version['revision']['id'], str(self.rev.id))

    def test_versions_list_filter_worker(self):
        self.client.force_login(self.user)
        worker_2 = Worker.objects.create(
            repository=self.repo,
            name='Worker 2',
            slug='worker_2',
            type=MLToolType.Classifier
        )
        version_2 = WorkerVersion.objects.create(
            worker=worker_2,
            revision=self.rev,
            configuration={"test": "test2"}
        )

        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:worker-versions', kwargs={'pk': str(worker_2.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data), 1)
        version = data[0]
        self.assertEqual(version['id'], str(version_2.id))

    # Tests on post for WorkerVersionList
    def test_versions_post_requires_login(self):
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}),
            data={'revision': str(self.rev2.id), 'configuration': {"test": "test2"}}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_versions_post_invalid_repository(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}),
            data={'revision': str(self.rev2.id), 'configuration': {"test": "test2"}}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_versions_post_create_worker_version(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}),
            data={'revision': str(self.rev2.id), 'configuration': {"test": "test2"}}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertNotEqual(data['id'], str(self.version_1.id))
        self.assertEqual(data['configuration'], {"test": "test2"})
        self.assertEqual(data['revision']['id'], str(self.rev2.id))
        self.assertEqual(data['state'], 'created')

    def test_versions_post_available_requires_docker_image(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}),
            data={
                'revision': str(self.rev2.id),
                'configuration': {"test": "test2"},
                'state': 'available',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_versions_post_return_existing_worker_version(self):
        self.client.force_login(self.user)
        # A worker version already exists for this worker and this revision
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}),
            data={'revision': str(self.rev.id), 'configuration': {"test": "test1"}}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.version_1.id))
        self.assertEqual(data['configuration'], {"test": 42})
        self.assertEqual(data['revision']['id'], str(self.rev.id))
        self.assertEqual(data['state'], 'available')

    def test_versions_post_empty(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'revision': ['This field is required.'],
            'configuration': ['This field is required.']
        })

    def test_retrieve_version_requires_login(self):
        response = self.client.get(reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_version_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:version-retrieve', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_version(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        revision = data.pop('revision')
        self.assertEqual(revision['id'], str(self.rev.id))
        worker = data.pop('worker')
        self.assertEqual(worker['id'], str(self.worker_1.id))
        self.maxDiff = None
        self.assertDictEqual(data, {
            'id': str(self.version_1.id),
            'configuration': {"test": 42},
            'docker_image': str(self.version_1.docker_image.id),
            'docker_image_iid': None,
            'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
            'state': 'available',
        })

    def test_update_version_requires_login(self):
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': '12341234-1234-1234-1234-123412341234',
                'state': 'error'
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_version_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': '12341234-1234-1234-1234-123412341234',
                'state': 'error'
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_worker_version(self):
        """
        Update worker version artifact, configuration and state
        """
        docker_image = self.version_1.docker_image
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()

        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': str(docker_image.id),
                'state': 'error'
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.version_1.id))
        self.version_1.refresh_from_db()
        self.assertEqual(data['configuration'], {"test": "test2"})
        self.assertEqual(data['docker_image'], str(docker_image.id))
        self.assertIsNone(data['docker_image_iid'])
        self.assertEqual(data['state'], 'error')

    def test_update_version_valid(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()

        workflow = Workflow.objects.create(recipe=RECIPE)
        workflow.start()
        task = workflow.tasks.get(slug='first')
        artifact = task.artifacts.create(
            path='path/to/file.json',
            size=100,
            content_type='application/json',
        )

        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': str(artifact.id),
                'docker_image_iid': 'sha256:deadbeef',
                'state': 'available'
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.version_1.id))
        self.assertEqual(data['configuration'], {"test": "test2"})
        self.assertEqual(data['docker_image'], str(artifact.id))
        self.assertEqual(data['docker_image_iid'], 'sha256:deadbeef')
        self.assertEqual(data['state'], 'available')

    def test_update_version_available_requires_docker_image(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()

        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={'state': 'available'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_update_version_available_docker_image_not_null(self):
        workflow = Workflow.objects.create(recipe=RECIPE)
        workflow.start()
        task = workflow.tasks.get(slug='first')
        self.version_1.docker_image = task.artifacts.create(
            path='path/to/file.json',
            size=100,
            content_type='application/json',
        )
        self.version_1.state = WorkerVersionState.Available
        self.version_1.save()

        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={'docker_image': None},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_update_version_available_docker_image_iid_not_null(self):
        workflow = Workflow.objects.create(recipe=RECIPE)
        workflow.start()
        task = workflow.tasks.get(slug='first')
        self.version_1.docker_image = task.artifacts.create(
            path='path/to/file.json',
            size=100,
            content_type='application/json',
        )
        self.version_1.docker_image_iid = 'something:latest'
        self.version_1.state = WorkerVersionState.Available
        self.version_1.save()

        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={'docker_image_iid': None},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_property_docker_image_name(self):
        self.assertEqual(self.version_1.docker_image_name, f'my_repo.fake/workers/worker/reco:{self.version_1.id}')

    def test_property_docker_image_name_explicit(self):
        version_2 = WorkerVersion.objects.create(
            id="12341234-1234-1234-1234-123412341234",
            worker=self.worker_1,
            revision=self.rev2,
            configuration={"test": "test1"}
        )
        self.assertEqual(
            version_2.docker_image_name,
            'my_repo.fake/workers/worker/reco:12341234-1234-1234-1234-123412341234'
        )

    def test_property_slug(self):
        self.assertEqual(self.version_1.slug, f'reco_{str(self.version_1.id)[0:6]}')

    def test_property_slug_explicit(self):
        version_2 = WorkerVersion.objects.create(
            id="01234567-89ab-cdef-fedc-ba9876543210",
            worker=self.worker_1,
            revision=self.rev2,
            configuration={"test": "test1"}
        )
        self.assertEqual(version_2.slug, 'reco_012345')

    def test_property_docker_command_empty(self):
        self.assertEqual(self.version_1.docker_command, None)

    def test_property_docker_command_not_empty(self):
        self.version_1.configuration = {'docker': {'command': 'mysupercommand'}}
        self.assertEqual(self.version_1.docker_command, 'mysupercommand')
        self.version_1.configuration = {"test": "test1"}
