from django.contrib import admin
from django import forms
from django_admin_hstore_widget.forms import HStoreFormField
from arkindex.documents.models import \
    Corpus, Element, ElementType, Transcription, MetaData, AllowedMetaData, MLClass, \
    Classification, DataSource, Entity, EntityRole, EntityLink
from enumfields.admin import EnumFieldListFilter


class ElementTypeInline(admin.TabularInline):
    model = ElementType
    prepopulated_fields = {'slug': ('display_name', )}


class CorpusAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'public', 'repository')
    search_fields = ('name', )
    inlines = (ElementTypeInline, )

    def has_delete_permission(self, request, obj=None):
        # Require everyone to use the asynchronous corpus deletion
        return False


class DataSourceAdmin(admin.ModelAdmin):
    list_display = ('id', 'type', 'slug', 'revision', 'internal')
    list_filter = [('type', EnumFieldListFilter), 'internal']
    readonly_fields = ('id', )


class ClassificationInline(admin.TabularInline):
    model = Classification
    readonly_fields = ('confidence', 'high_confidence', )


class AllowedMetaDataAdmin(admin.ModelAdmin):
    list_display = ('id', 'corpus', 'type', 'name')
    readonly_fields = ('id', )


class MetaDataAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'type', )
    list_filter = [('type', EnumFieldListFilter), ]
    readonly_fields = ('id', 'revision')
    raw_id_fields = ('element', 'entity', )


class MetaDataInline(admin.TabularInline):
    model = MetaData
    raw_id_fields = ('entity', )


class ElementAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'type', 'corpus', )
    list_filter = ['type__slug', 'corpus']
    fields = ('id', 'type', 'name', 'corpus')
    readonly_fields = ('id', )
    search_fields = ('name', )
    inlines = (MetaDataInline, ClassificationInline)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == 'type':
            # Only display types available on the element's corpus
            # It would make no sense to pick a type from another corpus
            # Display all elements on creation, as we do not know yet the corpus
            element_id = request.resolver_match.kwargs.get('object_id')
            if element_id:
                element = self.get_object(request, element_id)
                kwargs['queryset'] = ElementType.objects.filter(corpus=element.corpus).order_by('display_name')

        return super().formfield_for_foreignkey(db_field, request, **kwargs)


class TranscriptionAdmin(admin.ModelAdmin):
    list_display = ('id', 'text', 'score', 'element', )
    list_filter = [('type', EnumFieldListFilter), 'source']
    fields = ('id', 'text', 'score', 'element', 'source', )
    readonly_fields = ('id', )
    raw_id_fields = ('element', )


class MLClassAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'corpus')
    list_filter = ('corpus',)
    search_fields = ('name',)
    fields = ('name', 'corpus')


class EntityMetaForm(forms.ModelForm):
    metas = HStoreFormField()


class EntityLinkInLine(admin.TabularInline):
    model = EntityLink
    fk_name = 'parent'
    raw_id_fields = ('child', )


class EntityAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'type')
    list_filter = [('type', EnumFieldListFilter), 'corpus']
    readonly_fields = ('id', )
    search_fields = ('name', )
    inlines = (EntityLinkInLine, )
    form = EntityMetaForm


class EntityRoleAdmin(admin.ModelAdmin):
    list_display = ('id', 'corpus', 'parent_name', 'child_name')
    list_filter = ('corpus', )
    readonly_fields = ('id', )
    ordering = ('corpus', 'parent_name', 'child_name')


admin.site.register(Corpus, CorpusAdmin)
admin.site.register(DataSource, DataSourceAdmin)
admin.site.register(Element, ElementAdmin)
admin.site.register(Transcription, TranscriptionAdmin)
admin.site.register(MLClass, MLClassAdmin)
admin.site.register(MetaData, MetaDataAdmin)
admin.site.register(Entity, EntityAdmin)
admin.site.register(EntityRole, EntityRoleAdmin)
admin.site.register(AllowedMetaData, AllowedMetaDataAdmin)
