from uuid import UUID
from django.core.exceptions import ValidationError
from django.conf import settings
from django.db.models import Q
from elasticsearch.exceptions import NotFoundError
from rest_framework import serializers, status
from rest_framework.generics import \
    ListAPIView, ListCreateAPIView, RetrieveAPIView, CreateAPIView, RetrieveUpdateDestroyAPIView
from rest_framework.exceptions import NotFound
from rest_framework.response import Response
from arkindex.documents.models import \
    Corpus, Right, Element, Transcription, TranscriptionEntity, Entity, EntityLink, EntityRole
from arkindex.documents.serializers.entities import (
    EntityRoleSerializer, EntitySerializer, EntityCreateSerializer, EntityLinkCreateSerializer,
    TranscriptionEntitySerializer, TranscriptionEntityDetailsSerializer, EntityLinkSerializer,
)
from arkindex.documents.serializers.elements import ElementSlimSerializer, ElementEntitiesSerializer
from arkindex.dataimport.models import WorkerVersion
from arkindex.project.mixins import CorpusACLMixin
from arkindex.project.permissions import IsVerified
from arkindex.project.triggers import reindex_start
from arkindex.project.elastic import ESEntity


class CorpusRoles(CorpusACLMixin, ListCreateAPIView):
    """
    List all roles in a corpus
    """
    serializer_class = EntityRoleSerializer
    openapi_overrides = {
        'tags': ['entities']
    }

    def get_queryset(self):
        return EntityRole.objects \
            .filter(corpus=self.get_corpus(self.kwargs['pk'])) \
            .order_by('parent_name', 'child_name')

    def perform_create(self, serializer):
        data = self.request.data
        if EntityRole.objects.filter(
            parent_name=data['parent_name'],
            child_name=data['child_name'],
            parent_type=data['parent_type'],
            child_type=data['child_type'],
            corpus_id=self.request.parser_context['kwargs']['pk']
        ).exists():
            raise serializers.ValidationError({
                'corpus': ['Role already exists in this corpus'],
                'id': self.request.parser_context['kwargs']['pk']
            })
        super().perform_create(serializer)


class EntityDetails(RetrieveUpdateDestroyAPIView):
    """
    Get all information about an entity
    """
    serializer_class = EntitySerializer
    openapi_overrides = {
        'tags': ['entities'],
    }

    def get_queryset(self):
        return Entity.objects \
            .filter(corpus__in=Corpus.objects.readable(self.request.user)) \
            .prefetch_related(
                'parents__role',
                'children__role',
                'parents__child',
                'parents__parent',
                'children__parent',
                'children__child',
                'corpus'
            )

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if request.method not in ('GET', 'OPTIONS') and Right.Write not in obj.corpus.get_acl_rights(request.user):
            self.permission_denied(request, message='You do not have write access to this corpus.')

    def perform_destroy(self, instance):
        if settings.ARKINDEX_FEATURES['search']:
            # Try to delete indexed entity if possible
            try:
                es_entity = ESEntity.get(id=instance.id.hex)
            except NotFoundError:
                pass
            else:
                es_entity.delete()
        instance.delete()


class EntityElements(ListAPIView):
    """
    Get all elements that have a link with the entity
    """
    serializer_class = ElementSlimSerializer
    openapi_overrides = {
        'operationId': 'ListEntityElements',
        'security': [],
        'tags': ['entities'],
    }

    def get_queryset(self):
        pk = self.kwargs['pk']
        metadata_elements = Element.objects \
            .filter(
                corpus__in=Corpus.objects.readable(self.request.user),
                metadatas__entity_id=pk
            ) \
            .select_related('type', 'corpus') \
            .prefetch_related('metadatas__entity', 'metadatas__revision', 'zone__image__server')
        transcription_elements = Element.objects \
            .filter(
                corpus__in=Corpus.objects.readable(self.request.user),
                transcriptions__transcription_entities__entity_id=pk
            ).select_related('type', 'corpus') \
            .prefetch_related('metadatas__entity', 'metadatas__revision', 'zone__image__server')
        return metadata_elements.union(transcription_elements) \
            .order_by('name', 'type')


class EntityCreate(CreateAPIView):
    """
    Create a new entity
    """
    permission_classes = (IsVerified, )
    serializer_class = EntityCreateSerializer
    openapi_overrides = {
        'operationId': 'CreateEntity',
        'tags': ['entities'],
    }

    def perform_create(self, serializer):
        name = serializer.validated_data['name']
        type = serializer.validated_data['type']
        corpus = serializer.validated_data['corpus']
        metas = serializer.validated_data['metas'] if 'metas' in serializer.data else None
        worker_version = serializer.validated_data['worker_version']
        return Entity.objects.create(
            name=name,
            type=type,
            corpus=corpus,
            metas=metas,
            worker_version=worker_version
        )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        entity = EntitySerializer(self.perform_create(serializer))
        # Index entity into Elasticsearch
        reindex_start(entity=entity.instance, entities=True)
        headers = self.get_success_headers(serializer.data)
        return Response(entity.data, status=status.HTTP_201_CREATED, headers=headers)


class EntityLinkCreate(CreateAPIView):
    """
    Create a new link between two entities with a role
    """
    permission_classes = (IsVerified, )
    serializer_class = EntityLinkCreateSerializer
    openapi_overrides = {
        'operationId': 'CreateEntityLink',
        'tags': ['entities'],
    }


class TranscriptionEntityCreate(CreateAPIView):
    """
    Link an existing Entity to a given transcription with its position
    """
    permission_classes = (IsVerified, )
    serializer_class = TranscriptionEntitySerializer
    openapi_overrides = {
        'operationId': 'CreateTranscriptionEntity',
        'tags': ['entities'],
    }

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if not self.request:
            # OpenAPI cannot pass request context
            return context

        # We need to use the default database to avoid stale read
        # when assigning a newly created entity on on a newly created transcription
        tr = Transcription.objects.using('default').filter(
            id=self.kwargs.get('pk'),
            element__corpus__in=Corpus.objects.writable(self.request.user)
        ).first()
        return {**context, 'transcription': tr}

    def perform_create(self, serializer):
        data = serializer.validated_data
        if TranscriptionEntity.objects.filter(**data).exists():
            raise serializers.ValidationError({
                '__all__': ['Entity is already linked to this transcription at this position']
            })
        try:
            TranscriptionEntity(**data).clean()
        except ValidationError as e:
            raise serializers.ValidationError(serializers.as_serializer_error(e))
        super().perform_create(serializer)


class TranscriptionEntities(ListAPIView):
    """
    List existing entities linked to a specific transcription
    """
    serializer_class = TranscriptionEntityDetailsSerializer
    openapi_overrides = {
        'operationId': 'ListTranscriptionEntities',
        'tags': ['entities'],
        'security': [],
        'parameters': [
            {
                'name': 'worker_version',
                'in': 'query',
                'description': 'Only include entities created by a specific worker version',
                'required': False,
                'schema': {
                    'type': 'string',
                    'format': 'uuid',
                }
            },
        ]
    }

    def get_queryset(self):
        filters = {
            'transcription__element__corpus__in': Corpus.objects.readable(self.request.user),
            'transcription_id': self.kwargs['pk'],
        }
        if 'worker_version' in self.request.query_params:
            try:
                worker_version_id = UUID(self.request.query_params['worker_version'])
            except (TypeError, ValueError):
                raise serializers.ValidationError({'worker_version': ['Invalid UUID']})

            if not WorkerVersion.objects.filter(id=worker_version_id).exists():
                raise serializers.ValidationError({'worker_version': ['This worker version does not exist.']})

            filters['worker_version_id'] = worker_version_id

        return TranscriptionEntity.objects \
            .filter(**filters) \
            .order_by('offset') \
            .prefetch_related('entity')


class ElementEntities(RetrieveAPIView):
    """
    List all entities linked to an element's transcriptions and metadata
    """
    serializer_class = ElementEntitiesSerializer
    openapi_overrides = {
        'operationId': 'ListElementEntities',
        'tags': ['entities'],
        'security': [],
        'parameters': [
            {
                'name': 'worker_version',
                'in': 'query',
                'description': 'Only include entities created by a specific worker version',
                'required': False,
                'schema': {
                    'type': 'string',
                    'format': 'uuid',
                }
            },
        ]
    }

    def get_queryset(self):
        corpora = Corpus.objects.readable(self.request.user)
        return Element.objects.filter(corpus__in=corpora).select_related('type').only('id', 'type__slug', 'name')


class ElementLinks(ListAPIView):
    """
    List all links where parent and child are linked to the element
    """
    serializer_class = EntityLinkSerializer
    openapi_overrides = {
        'operationId': 'ListElementLinks',
        'tags': ['entities'],
    }

    def get_queryset(self):
        try:
            element = Element.objects.get(id=self.kwargs['pk'])
        except Element.DoesNotExist:
            raise NotFound
        entities = Entity.objects.filter(Q(metadatas__element_id=element.id) | Q(transcriptions__element_id=element.id))
        return EntityLink.objects.filter(parent__in=entities, child__in=entities) \
            .prefetch_related('role', 'child', 'parent').order_by('parent__name')
