from django.urls import reverse
from rest_framework import status
from arkindex.project.tests import FixtureAPITestCase
from arkindex.documents.models import Element, ElementPath
from arkindex.dataimport.models import WorkerVersion


class TestBulkElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.payload = {
            'worker_version': str(cls.worker_version.id),
            'elements': [
                {
                    'name': 'A',
                    'type': 'act',
                    'polygon': [
                        [0, 0],
                        [0, 1],
                        [1, 1],
                        [1, 0],
                        [0, 0],
                    ]
                },
                {
                    'name': 'B',
                    'type': 'surface',
                    # Reuses the same zone
                    'polygon': [
                        [0, 0],
                        [0, 1],
                        [1, 1],
                        [1, 0],
                        [0, 0],
                    ]
                },
                {
                    'name': 'C',
                    'type': 'surface',
                    'polygon': [
                        [0, 0],
                        [0, 9],
                        [9, 9],
                        [9, 0],
                        [0, 0],
                    ]
                }
            ]
        }

    def setUp(self):
        self.element = self.corpus.elements.get(name='Volume 2, page 1v')

    def test_bulk_create_requires_login(self):
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
            data=self.payload,
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_bulk_create_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
            data=self.payload,
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_bulk_create_writable_corpus(self):
        self.user.corpus_right.filter(corpus=self.corpus).update(can_write=False, can_admin=False)
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
            data=self.payload,
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_bulk_create_requires_parent_zone(self):
        self.element.zone.delete()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
            data={'elements': []},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'element': ['Cannot create child elements on an element without a zone.']
        })

    def test_bulk_create_requires_elements(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
            data={'elements': []},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'elements': {'non_field_errors': ['This list may not be empty.']}})

    def test_bulk_create_missing_types(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
                data={'elements': [
                    {
                        'name': 'nope',
                        'type': 'nope',
                        'polygon': [[0, 0], [0, 42], [13, 37]],
                    }
                ]},
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'non_field_errors': ["Element types with slugs nope do not exist in the parent element's corpus"]
        })

    def test_bulk_create(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
                data=self.payload,
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        element_path = self.element.paths.get()
        a, b, c = Element \
            .objects \
            .get_descending(self.element.id) \
            .filter(type__slug__in=('act', 'surface')) \
            .order_by('name') \
            .select_related('zone')
        a_path, b_path, c_path = a.paths.get(), b.paths.get(), c.paths.get()

        self.assertListEqual(
            response.json(),
            [
                {'id': str(a.id)},
                {'id': str(b.id)},
                {'id': str(c.id)}
            ]
        )

        self.assertEqual(a.name, 'A')
        self.assertEqual(b.name, 'B')
        self.assertEqual(c.name, 'C')
        self.assertEqual(a.type.slug, 'act')
        self.assertEqual(b.type.slug, 'surface')
        self.assertEqual(c.type.slug, 'surface')
        self.assertEqual(a.worker_version, self.worker_version)
        self.assertEqual(b.worker_version, self.worker_version)
        self.assertEqual(c.worker_version, self.worker_version)
        self.assertEqual(a.zone.image_id, self.element.zone.image_id)
        self.assertEqual(b.zone.image_id, self.element.zone.image_id)
        self.assertEqual(c.zone.image_id, self.element.zone.image_id)
        self.assertTupleEqual(a.zone.polygon.coords, ((0, 0), (0, 1), (1, 1), (1, 0), (0, 0)))
        self.assertTupleEqual(b.zone.polygon.coords, ((0, 0), (0, 1), (1, 1), (1, 0), (0, 0)))
        self.assertTupleEqual(c.zone.polygon.coords, ((0, 0), (0, 9), (9, 9), (9, 0), (0, 0)))
        # Both elements reuse the same polygon
        self.assertEqual(a.zone, b.zone)

        self.assertListEqual(a_path.path, element_path.path + [self.element.id])
        self.assertListEqual(b_path.path, element_path.path + [self.element.id])
        self.assertListEqual(c_path.path, element_path.path + [self.element.id])
        self.assertEqual(a_path.ordering, 0)
        self.assertEqual(b_path.ordering, 0)
        self.assertEqual(c_path.ordering, 1)

    def test_bulk_create_multiple_parent_paths(self):
        parent = self.corpus.elements.create(
            name='Parent 2',
            type=self.corpus.types.first(),
        )
        self.element.add_parent(parent)
        element_path1, element_path2 = self.element.paths.all()

        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
                data=self.payload,
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        a, b, c = Element \
            .objects \
            .get_descending(self.element.id) \
            .filter(type__slug__in=('act', 'surface')) \
            .order_by('name') \
            .select_related('zone')

        a_path1, a_path2 = a.paths.all()
        b_path1, b_path2 = b.paths.all()
        c_path1, c_path2 = c.paths.all()

        self.assertListEqual(a_path1.path, element_path1.path + [self.element.id])
        self.assertListEqual(a_path2.path, element_path2.path + [self.element.id])
        self.assertListEqual(b_path1.path, element_path1.path + [self.element.id])
        self.assertListEqual(b_path2.path, element_path2.path + [self.element.id])
        self.assertListEqual(c_path1.path, element_path1.path + [self.element.id])
        self.assertListEqual(c_path2.path, element_path2.path + [self.element.id])

        self.assertEqual(a_path1.ordering, 0)
        self.assertEqual(a_path2.ordering, 0)
        self.assertEqual(b_path1.ordering, 0)
        self.assertEqual(b_path2.ordering, 0)
        self.assertEqual(c_path1.ordering, 1)
        self.assertEqual(c_path2.ordering, 1)

    def test_bulk_create_existing_zone(self):
        polygon = [[680, 296], [681, 359], [2252, 343], [2251, 280], [680, 296]]

        # Create a specific zone on an element
        zone = self.element.zone.image.zones.create(polygon=polygon)

        # Try to create a sub element on that zone
        self.client.force_login(self.user)
        with self.assertNumQueries(12):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
                data={
                    'worker_version': str(self.worker_version.id),
                    'elements': [
                        {
                            'name': 'dupe',
                            'type': 'surface',
                            'polygon': polygon
                        },
                    ]
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        # Check the created element uses the pre-existing zone
        data = response.json()
        self.assertEqual(len(data), 1)
        elt = Element.objects.get(pk=data[0]["id"])
        self.assertEqual(elt.zone, zone)

    def test_bulk_create_top_level(self):

        # Create a top level page, without any path
        element = self.corpus.elements.create(
            name="Top level page",
            type=self.corpus.types.get(slug='page'),
            zone=self.element.zone,  # reuse existing zone/image
        )
        self.assertFalse(element.paths.exists())

        # Try to create a sub element on that zone
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(element.id)}),
                data={
                    'worker_version': str(self.worker_version.id),
                    'elements': [
                        {
                            'name': f'surface {i}',
                            'type': 'surface',
                            'polygon': [
                                [i * 100, i * 100],
                                [(i + 1) * 100, i * 100],
                                [(i + 1) * 100, (i + 1) * 100],
                                [i * 100, (i + 1) * 100],
                            ]
                        }
                        for i in range(1, 5)
                    ]
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        # Check the created elements are children of element
        data = response.json()
        self.assertEqual(len(data), 4)
        paths = ElementPath.objects.filter(element_id__in=[elt['id'] for elt in data], path=[element.id])
        self.assertEqual(paths.count(), 4)
