from django.urls import path
from django.views.generic.base import RedirectView
from django.views.decorators.cache import cache_page
from django.conf import settings
from rest_framework.schemas import get_schema_view
from arkindex.project.openapi import SchemaGenerator

from arkindex.documents.api.elements import (
    CorpusElements, ElementsCreate, ElementBulkCreate, ElementRetrieve, ElementTranscriptions, ElementNeighbors,
    ElementParent, ElementParents, ElementChildren, ElementMetadata, MetadataEdit, ManageSelection,
    CorpusList, CorpusRetrieve, CorpusAllowedMetaData, ElementTypeCreate, ElementTypeUpdate, DeprecatedElementsList
)
from arkindex.documents.api.search import ElementSearch, EntitySearch
from arkindex.documents.api.ml import (
    ClassificationCreate, ClassificationValidate, ClassificationReject, ClassificationBulk, ElementTranscriptionsBulk,
    TranscriptionCreate, TranscriptionBulk, ElementMLStats, CorpusMLStats,
    ManageClassificationsSelection, MLClassList, CorpusMLClassList, TranscriptionEdit
)
from arkindex.documents.api.entities import (
    CorpusRoles, EntityDetails, EntityCreate, EntityLinkCreate, EntityElements,
    TranscriptionEntityCreate, TranscriptionEntities, ElementEntities, ElementLinks
)
from arkindex.documents.api.iiif import FolderManifest, ElementAnnotationList, TranscriptionSearchAnnotationList
from arkindex.documents.api.admin import ReindexStart
from arkindex.dataimport.api import (
    DataImportsList, DataImportDetails, DataImportRetry, StartProcess, DataImportElements,
    DataFileList, DataFileRetrieve, DataFileUpload, DataImportFromFiles,
    RepositoryList, RepositoryRetrieve, DataFileCreate,
    GitRepositoryImportHook, AvailableRepositoriesList, RevisionRetrieve,
    MLToolList, CorpusWorkflow, WorkerList, WorkerVersionList, WorkerVersionRetrieve,
    WorkerRunList, WorkerRunDetails, ImportTranskribus, WorkerRetrieve, ListProcessElements
)
from arkindex.images.api import ImageCreate, IIIFURLCreate, IIIFInformationCreate, ImageRetrieve, ImageElements
from arkindex.users.api import (
    ProvidersList, CredentialsList, CredentialsRetrieve, OAuthSignIn, OAuthRetry, OAuthCallback,
    UserRetrieve, UserCreate, UserEmailLogin, UserEmailVerification, PasswordReset, PasswordResetConfirm,
    UpdateUserTranskribus
)

# Cache the OpenAPI schema view for a day
schema_view = cache_page(86400, key_prefix=settings.VERSION)(get_schema_view(generator_class=SchemaGenerator))

api = [

    # Elements
    path('elements/', DeprecatedElementsList.as_view(), name='elements'),
    path('elements/<uuid:pk>/neighbors/', ElementNeighbors.as_view(), name='elements-neighbors'),
    path('elements/<uuid:pk>/parents/', ElementParents.as_view(), name='elements-parents'),
    path('elements/<uuid:pk>/children/', ElementChildren.as_view(), name='elements-children'),
    path('elements/selection/', ManageSelection.as_view(), name='elements-selection'),
    path('elements/create/', ElementsCreate.as_view(), name='elements-create'),
    path('elements/type/', ElementTypeCreate.as_view(), name='element-type-create'),
    path('elements/type/<uuid:pk>/', ElementTypeUpdate.as_view(), name='element-type'),
    path('element/<uuid:pk>/', ElementRetrieve.as_view(), name='element-retrieve'),
    path('element/<uuid:pk>/children/bulk/', ElementBulkCreate.as_view(), name='elements-bulk-create'),
    path('element/<uuid:pk>/metadata/', ElementMetadata.as_view(), name='element-metadata'),
    path('element/<uuid:pk>/entities/', ElementEntities.as_view(), name='element-entities'),
    path('element/<uuid:pk>/links/', ElementLinks.as_view(), name='element-links'),
    path('element/<uuid:pk>/transcription/', TranscriptionCreate.as_view(), name='transcription-create'),
    path('element/<uuid:pk>/transcriptions/', ElementTranscriptions.as_view(), name='element-transcriptions'),
    path(
        'element/<uuid:pk>/transcriptions/bulk/',
        ElementTranscriptionsBulk.as_view(),
        name='element-transcriptions-bulk'
    ),
    path('element/<uuid:pk>/ml-stats/', ElementMLStats.as_view(), name='element-ml-stats'),
    path('element/<uuid:child>/parent/<uuid:parent>/', ElementParent.as_view(), name='element-parent'),

    # Corpora
    path('corpus/', CorpusList.as_view(), name='corpus'),
    path('corpus/<uuid:pk>/', CorpusRetrieve.as_view(), name='corpus-retrieve'),
    # `corpus` and not `pk` so that OpenAPI uses a `corpus` parameter
    path('corpus/<uuid:corpus>/elements/', CorpusElements.as_view(), name='corpus-elements'),
    path('corpus/<uuid:pk>/classes/', CorpusMLClassList.as_view(), name='corpus-classes'),
    path('corpus/<uuid:pk>/roles/', CorpusRoles.as_view(), name='corpus-roles'),
    path('corpus/<uuid:pk>/ml-stats/', CorpusMLStats.as_view(), name='corpus-ml-stats'),
    path('corpus/<uuid:pk>/allowed-metadata/', CorpusAllowedMetaData.as_view(), name='corpus-allowed-metadata'),

    # Moderation
    path('ml-classes/', MLClassList.as_view(), name='mlclass-list'),
    path('classifications/', ClassificationCreate.as_view(), name='classification-create'),
    path('classifications/<uuid:pk>/validate/', ClassificationValidate.as_view(), name='classification-validate'),
    path('classifications/<uuid:pk>/reject/', ClassificationReject.as_view(), name='classification-reject'),
    path('classifications/selection/', ManageClassificationsSelection.as_view(), name='classification-selection'),

    # Manifests
    path('iiif/<uuid:pk>/manifest/', FolderManifest.as_view(), name='folder-manifest'),
    path('iiif/<uuid:pk>/list/transcriptions/', ElementAnnotationList.as_view(), name='element-annotation-list'),
    path('iiif/<uuid:pk>/search/', TranscriptionSearchAnnotationList.as_view(), name='iiif-search'),

    # Placeholder URLs for IIIF IDs
    path(
        'iiif/<uuid:pk>/sequence/',
        RedirectView.as_view(pattern_name='api:folder-manifest', permanent=False),
        name='iiif-sequence',
    ),
    path(
        'iiif/<uuid:pk>/canvas/',
        RedirectView.as_view(pattern_name='api:element-annotation-list', permanent=False),
        name='iiif-canvas',
    ),
    path(
        'iiif/<uuid:page_pk>/annotation/<uuid:transcription_pk>/',
        RedirectView.as_view(pattern_name='api:element-annotation-list', permanent=False),
        name='transcription-annotation',
    ),

    # Search engines
    path('elements/search/', ElementSearch.as_view(), name='element-search'),
    path('entity/search/', EntitySearch.as_view(), name='entity-search'),

    # Ingest transcriptions
    path('transcription/<uuid:pk>/', TranscriptionEdit.as_view(), name='transcription-edit'),
    path('transcription/bulk/', TranscriptionBulk.as_view(), name='transcription-bulk'),
    path('classification/bulk/', ClassificationBulk.as_view(), name='classification-bulk'),

    # Ingest and retrieve entities
    path('entity/', EntityCreate.as_view(), name='entity-create'),
    path('entity/link/', EntityLinkCreate.as_view(), name='entity-link-create'),
    path('entity/<uuid:pk>/', EntityDetails.as_view(), name='entity-details'),
    path('entity/<uuid:pk>/elements/', EntityElements.as_view(), name='entity-elements'),
    path('transcription/<uuid:pk>/entity/', TranscriptionEntityCreate.as_view(), name='transcription-entity-create'),
    path('transcription/<uuid:pk>/entities/', TranscriptionEntities.as_view(), name='transcription-entities'),

    # Ingest metadata
    path('metadata/<uuid:pk>/', MetadataEdit.as_view(), name='metadata-edit'),

    # Git import workflows
    path('imports/repos/', RepositoryList.as_view(), name='repository-list'),
    path('imports/repos/<uuid:pk>/', RepositoryRetrieve.as_view(), name='repository-retrieve'),
    path('imports/repos/search/<uuid:pk>/', AvailableRepositoriesList.as_view(), name='available-repositories'),
    path('imports/hook/<uuid:pk>/', GitRepositoryImportHook.as_view(), name='import-hook'),
    path('imports/revisions/<uuid:pk>/', RevisionRetrieve.as_view(), name='revision-retrieve'),

    # Workers
    path('imports/repos/<uuid:pk>/workers/', WorkerList.as_view(), name='repository-workers'),
    path('workers/<uuid:pk>/', WorkerRetrieve.as_view(), name='worker-retrieve'),
    path('workers/<uuid:pk>/versions/', WorkerVersionList.as_view(), name='worker-versions'),
    path('workers/versions/<uuid:pk>/', WorkerVersionRetrieve.as_view(), name='version-retrieve'),

    # Import workflows
    path('imports/', DataImportsList.as_view(), name='import-list'),
    path('imports/fromfiles/', DataImportFromFiles.as_view(), name='import-from-files'),
    path('imports/mltools/', MLToolList.as_view(), name='ml-tool-list'),
    path('imports/transkribus/', ImportTranskribus.as_view(), name='import-transkribus'),
    path('imports/<uuid:pk>/', DataImportDetails.as_view(), name='import-details'),
    path('imports/<uuid:pk>/elements/', DataImportElements.as_view(), name='import-elements'),
    path('imports/<uuid:pk>/retry/', DataImportRetry.as_view(), name='import-retry'),
    path('imports/<uuid:pk>/start/', StartProcess.as_view(), name='process-start'),
    path('imports/files/<uuid:pk>/', DataFileList.as_view(), name='file-list'),
    path('imports/files/create/', DataFileCreate.as_view(), name='file-create'),
    path('imports/file/<uuid:pk>/', DataFileRetrieve.as_view(), name='file-retrieve'),
    path('imports/upload/<uuid:pk>/', DataFileUpload.as_view(), name='file-upload'),
    path('imports/corpus/', CorpusWorkflow.as_view(), name='corpus-workflow'),
    path('imports/<uuid:pk>/workers/', WorkerRunList.as_view(), name='worker-run-list'),
    path('imports/workers/<uuid:pk>/', WorkerRunDetails.as_view(), name='worker-run-details'),
    path('process/<uuid:pk>/elements/', ListProcessElements.as_view(), name='process-elements-list'),

    # Image management
    path('image/', ImageCreate.as_view(), name='image-create'),
    path('image/iiif/url/', IIIFURLCreate.as_view(), name='iiif-url-create'),
    path('image/iiif/info/', IIIFInformationCreate.as_view(), name='iiif-info-create'),
    path('image/<uuid:pk>/', ImageRetrieve.as_view(), name='image-retrieve'),
    path('image/<uuid:pk>/elements/', ImageElements.as_view(), name='image-elements'),

    # Manage OAuth integrations
    path('oauth/providers/', ProvidersList.as_view(), name='providers-list'),
    path('oauth/providers/<provider>/signin/', OAuthSignIn.as_view(), name='oauth-signin'),
    path('oauth/providers/<provider>/callback/', OAuthCallback.as_view(), name='oauth-callback'),
    path('oauth/credentials/', CredentialsList.as_view(), name='credentials-list'),
    path('oauth/credentials/<uuid:pk>/', CredentialsRetrieve.as_view(), name='credentials-retrieve'),
    path('oauth/credentials/<uuid:pk>/retry/', OAuthRetry.as_view(), name='oauth-retry'),

    # Authentication
    path('user/', UserRetrieve.as_view(), name='user-retrieve'),
    path('user/new/', UserCreate.as_view(), name='user-new'),
    path('user/login/', UserEmailLogin.as_view(), name='user-login'),
    path('user/token/', UserEmailVerification.as_view(), name='user-token'),
    path('user/password-reset/', PasswordReset.as_view(), name='password-reset'),
    path('user/password-reset/confirm/', PasswordResetConfirm.as_view(), name='password-reset-confirm'),
    path('user/transkribus/', UpdateUserTranskribus.as_view(), name='user-transkribus'),

    # Management tools
    path('reindex/', ReindexStart.as_view(), name='reindex-start'),

    # OpenAPI Schema
    path('openapi/', schema_view, name='openapi-schema'),
]
